/*
IAX balancing proxy
Authors: Ben Smithrust, Stanislaw Pitucha

Copyright (c) 2007 Gradwell dot com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <assert.h>
#include <string.h>

#include "clients.h"
#include "log.h"
#include "global.h"

static const char rcsid[] = "iaxproxy-version: " __FILE__ " " VCS_VERSION;

#ifndef S_SPLINT_S
struct list_ptrs clients[256] = { [0 ... 255] = LIST_HEAD };
#else
struct list_ptrs clients[256] = { LIST_HEAD };
#endif

client_search client_find_and_count(struct in_addr addr, in_port_t port) {
	struct list_ptrs *iter;
	client *clt;
	int this_ip = 0;
	client_search cs = {.found = 0, .result = {0}};

	list_foreach(&clients[ntohl(addr.s_addr)&0xff], iter) {
		clt = list_entry(iter, client, list);
		if((clt->addr.sin_addr.s_addr == addr.s_addr) && (clt->addr.sin_port == port)) {
			cs.found = 1;
			cs.result.clt = clt;
			return cs;
		} else if (clt->addr.sin_addr.s_addr == addr.s_addr)
			this_ip++;
	}
	
	cs.result.ip_count = this_ip;
	return cs;
}

client* client_get(struct in_addr addr, in_port_t port) {
	client *clt;
	client_search cs;

	cs = client_find_and_count(addr, port);
	
	if (cs.found) {
		return cs.result.clt;
	} else if (cs.result.ip_count >= MAX_CLT_PER_IP) {
		LOG(WARNING, "Connection limit exceeded for %s", inet_ntoa(addr));
		return NULL;
	}

	// not found - create a new one
	clt = (client*) malloc(sizeof(client));
	assert(clt!=NULL);

	memset(&clt->addr, 0, sizeof (clt->addr));
	clt->addr.sin_family = AF_INET;
	clt->addr.sin_addr.s_addr = addr.s_addr;
	clt->addr.sin_port = port;
	snprintf(clt->ip_port, sizeof clt->ip_port, "%s:%d", inet_ntoa(clt->addr.sin_addr), ntohs(clt->addr.sin_port));
	clt->ref_count = 0;
	clt->callno_count = 0;

	memset(&clt->connections, 0, sizeof (clt->connections));

	LOG(DEBUG, "New connection from %s - has %d connections already", clt->ip_port, cs.result.ip_count);

	list_add_after(&clients[ntohl(addr.s_addr)&0xff], &clt->list);

	return clt;
}

/*
int client_remove(in_addr_t addr, in_port_t port) {
	struct list_ptrs *iter;
	client *clt;

	list_foreach(&(clients[ntohl(addr)&0xff]), iter) {
		clt = list_entry(iter, client, list);
		if((clt->addr.sin_addr.s_addr == addr) && (clt->addr.sin_port == port)) {
			list_remove(iter);
			return 0;
		}
	}
	return 1;
}
*/

int client_get_callno(client *clt) {
	int no;

	if (clt->callno_count >= MAX_CALLNO_PER_CLT) {
		LOG(WARNING, "client %s reached max callno %d", clt->ip_port, clt->callno_count);
		return (-1);
	}

	no = free_callno_get();
	if (no != -1) {
		LOG(DEBUG, "gave callno %d to client %s", no, clt->ip_port);
		clt->callno_count++;
	}
	return no;
}

void client_return_callno(client *clt, int no) {
	free_callno_return(no);
	clt->callno_count--;
	if (clt->callno_count < 0)
		CRASH("callno on client %s gone negative", clt->ip_port);
}

void client_safe_destroy(client* clt) {
	if (clt->ref_count == 0)
		client_destroy(clt);
}

void client_destroy(client* clt) {
	LOG(DEBUG, "destroying client: %s", clt->ip_port);
	// removing from list is always safe - it will not follow NULLs in prev/next
	list_remove(&clt->list);
	free(clt);
}

void
client_list_print(void) {
	int i;
	int total;
	list_ptrs *iter;
	client *clt;

	total = 0;
	for (i = 0; i < 256; i++) {
		list_foreach(&clients[i], iter) {
			clt = list_entry(iter, client, list);
			LOG(DEBUG, "client: %s:  ref_count=%d", clt->ip_port, clt->ref_count);
			total++;
		}
	}
	LOG(DEBUG, "%d clients", total);
}

void client_ref_up(client* clt) {
    clt->ref_count++;
}

void
client_ref_down_destroy(client *clt) {
	client_ref_down(clt);
	if (clt->ref_count == 0)
		client_destroy(clt);
}

void client_ref_down(client* clt) {
	LOG(DEBUG, "client: %s pre ref_down, ref=%i", clt->ip_port, clt->ref_count);
	if(clt->ref_count>0) {
		clt->ref_count--;
		if(clt->ref_count==0) {
			LOG(NOTICE, "client: %s is no longer used, ready to remove", clt->ip_port);
		}
	} else {
		CRASH("client: %s has refcount <= 0 and requested decrement", clt->ip_port);
	}
}

void client_destroy_all(void) {
	list_ptrs *iter;
	int i;
	int destroyed = 0;

	for(i=0;i<256;i++) {
		while((iter = clients[i].next)!=NULL) {
			client *clt;
			clt = list_entry(iter, client, list);
			list_remove(iter);
			free(clt);
			destroyed++;
		}
	}
	LOG(NOTICE, "destroyed %i clients", destroyed);
}

int client_count(void) {
	list_ptrs *iter;
	int i, active = 0;

	for(i=0;i<256;i++) {
		list_foreach(&clients[i], iter) {
			active++;
		}
	}

	return active;
}

