/*
IAX balancing proxy
Authors: Ben Smithrust, Stanislaw Pitucha

Copyright (c) 2007 Gradwell dot com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#define _GNU_SOURCE

#include <assert.h>
#include <fcntl.h>
#include <string.h>

#include "clients.h"
#include "clients.h"
#include "connections.h"
#include "global.h"
#include "iax.h"
#include "log.h"
#include "servers.h"
#include "iaxproxy.h"

int master_sock;
static time_t last_poke_any = 0;

static void send_poke(send_command_t send_command, server *srv, time_t now) {
	struct iax_packet frame_full = {
		.dcallno=0, .ts=0,
		.oseqno=0, .iseqno=0,
		.type=AST_FRAME_IAX, .csub=IAX_TYPE_POKE
	};
	connection *conn;

	LOG(NOTICE, "Server at %s will be POKEd", srv->ip_string);

	conn = connection_create_from_balancer(srv);
	if (conn == NULL) {
		LOG(WARNING, "Failed to create connection to poke %s", srv->ip_string);
		return;
	}

	srv->last_poke = now;
	last_poke_any = now;

	ADDRESS_FULL_TO_SERVER((&frame_full), conn);
	conn->last_packet = now;

	if(send_command(master_sock, (char *)&frame_full, (socklen_t)sizeof frame_full, 0, (struct sockaddr*)&conn->srv->addr, sizeof(conn->srv->addr))==-1) {
		LOG(ERROR, "sendto() returned %s", strerror(errno));
	}
}

void check_servers_pokes(send_command_t send_command, struct timeval *now) {
	list_ptrs *iter_srv;

	list_foreach(&servers, iter_srv) {
		server *srv = list_entry(iter_srv, server, list);

		if(srv->enabled==0)
			continue;

		// host inactive for last POKE_INTERVAL secs
		if(now->tv_sec - srv->last_activity >= POKE_INTERVAL) {
			// check if server was active after last pong
			if(srv->last_pong < srv->last_activity)
				srv->last_pong=srv->last_activity;

			// check if server has unanswered poke & waiting timed out
			if((srv->last_poke > srv->last_pong) && ((now->tv_sec - srv->last_pong) > POKE_MAXWAIT)) {
				if(srv->not_responsive==0) {
					LOG(ERROR, "Server at %s stopped responding", srv->ip_string);
				}
				srv->not_responsive = 1;
			}

			// check if it's time to poke again
			if(now->tv_sec - srv->last_poke >= POKE_INTERVAL && now->tv_sec > last_poke_any) {
				send_poke(send_command, srv, now->tv_sec);
			}
		}
	}
}

// change packet into return INVAL packet
static void packet_create_inval(struct iax_packet *frame_full) {
	uint16_t scall, dcall;
	uint8_t iseq, oseq;

	scall = iax_get_scallno(frame_full);
	dcall = iax_get_dcallno(frame_full);
	frame_full->scallno=htons(dcall | 0x8000);
	frame_full->dcallno=htons(scall);

	iseq = frame_full->iseqno;
	oseq = frame_full->oseqno;
	frame_full->iseqno = oseq+1;
	frame_full->oseqno = iseq;

	frame_full->type = AST_FRAME_IAX;
	frame_full->csub = IAX_TYPE_INVAL;
}

// handling of closing packets
static void process_packet_closing(connection *conn, struct iax_packet *frame_full, int closing_party) {
	switch(frame_full->csub) {
		case IAX_TYPE_HANGUP:
		case IAX_TYPE_REJECT:
			// if it's hangup - set closing timestamp
			conn->closing=ntohl(frame_full->ts);
			conn->is_closing=1;
			break;

		case IAX_TYPE_ACK:
			// if it's ack - check if closing and if the timestamp is correct
			if(conn->is_closing!=0 && conn->closing==ntohl(frame_full->ts))
				connection_cleanup_check(conn, closing_party); // force close & cleanup
			break;
	}
}

static int close_if_received_inval(connection *conn, struct iax_packet *frame_full, int cause) {
	if(frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_INVAL) {
		LOG_CONN(WARNING, "Got INVAL - closing", conn);
		connection_cleanup_check(conn, cause);
		return 1;
	}
	return 0;
}

static void process_packet_poke(send_command_t send_command, client *clt, struct iax_packet *frame_full, size_t len, struct sockaddr_in *addr) {
	int32_t no;
	short x;

	LOG(NOTICE, "got POKE from %s - have to reply PONG", clt->ip_port);
	no = client_get_callno(clt);
	if(no==-1) {
		LOG(ERROR, "failed to get a callno");
		packet_create_inval(frame_full);
		len = sizeof *frame_full;
	} else {
		// there will be ack response
		ack_wait_add(clt, iax_get_scallno(frame_full), (uint16_t) no, ntohl(frame_full->ts));
		// add ref to client, so it will not disappear before ack
		client_ref_up(clt);
		// respond to POKE with PONG
		x = iax_get_scallno(frame_full);
		frame_full->dcallno = htons(x);
		// just to be sure it wont interfere with other connections
		frame_full->scallno = htons(((uint16_t) no) | 0x8000);
		frame_full->csub = IAX_TYPE_PONG;
		frame_full->iseqno++;
	}

	if(send_command(master_sock, (void *)frame_full, len, 0, (struct sockaddr*)addr, sizeof(*addr))==-1) {
		LOG(ERROR, "sendto() returned %s", strerror(errno));
	}
}

static void process_packet_from_server_full(send_command_t send_command, server *srv, struct iax_packet *frame_full, size_t len, struct sockaddr_in *addr, struct timeval *now) {
	connection *conn;

	// check if packet is big enough to analyze - ignore if not
	if(len<sizeof(struct iax_packet)) {
		LOG(WARNING, "bad packet from %s (too short for IAX) dropped", srv->ip_string);
		return;
	}

	// find corresponding connection
	conn = connection_get_by_srv(srv, iax_get_dcallno(frame_full), iax_get_scallno(frame_full));
	if(conn==NULL) { //unknown frame
		LOG(WARNING, "unknown server full frame %i, %i (%s) from %s, scallno: %i, dcallno: %i, iseq: %i, oseq: %i",
			frame_full->type, frame_full->csub,
			FRAME_NAME(frame_full),
			srv->ip_string,
			iax_get_scallno(frame_full), iax_get_dcallno(frame_full),
			frame_full->iseqno, frame_full->oseqno);
		// unknown call - return INVAL
		packet_create_inval(frame_full);
		if(send_command(master_sock, (void*)frame_full, sizeof (struct iax_packet), 0, (struct sockaddr*) addr, sizeof (struct sockaddr_in))<0) {
			err(1, "sending inval frame back to server");
		}
		return;
	}

	// update last packet receiving time
	(void) time(&conn->last_packet);

	// check / update the retransmissions
	retransmission_clear(RETR_DEST_SERVER, frame_full->iseqno - 1, &conn->retransmissions_list);

	// if that's POKE connection to server
	if(conn->clt == NULL) {
		LOG_CONN(VERBOSE_DEBUG, "server->balancer full frame %i, %i (%s) ",
			conn, frame_full->type, frame_full->csub,
			FRAME_NAME(frame_full));
		if(frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_PONG) {
			// TODO: check if packet corresponds to POKE - now just accept & ACK
			(void) time(&srv->last_pong);
			if(srv->not_responsive!=0)
				LOG(INFO, "Server %s is back up!", srv->ip_string);
			srv->not_responsive=0;

			ADDRESS_FULL_TO_SERVER(frame_full, conn);
			frame_full->csub = IAX_TYPE_ACK;
			if(send_command(master_sock, (void *)frame_full, len, 0, (struct sockaddr*)&conn->srv->addr, sizeof(conn->srv->addr))==-1) {
				LOG(ERROR, "sendto() returned %s", strerror(errno));
			}

			LOG(DEBUG, "Got PONG and returned ACK - server %s alive", srv->ip_string);

			connection_cleanup_check(conn, CLOSED_INTERNAL_IGNORE);
		}
	} else { // normal call
		if(conn->is_closing==0 || (frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_ACK)) {
			LOG_CONN(VERBOSE_DEBUG, "server full frame %i, %i (%s)",
				conn, frame_full->type, frame_full->csub,
				FRAME_NAME(frame_full));

			// save routing time to be printed when call finishes
			if (frame_full->type==AST_FRAME_IAX && (frame_full->csub==IAX_TYPE_ACCEPT || frame_full->csub==IAX_TYPE_REJECT))
				conn->setup_timestamp = ntohl(frame_full->ts);

			// rewrite destination callno  - was balancer_callno
			// rewrite source callno to balancer_callno
			ADDRESS_FULL_TO_CLIENT(frame_full, conn);

			if(send_command(master_sock, (void*)frame_full, len, 0, (struct sockaddr*)&conn->clt->addr, sizeof (struct sockaddr_in))<0) {
				err(1, "sending full frame to client");
			} else {
				if(conn->is_closing==0 && !(frame_full->type==AST_FRAME_IAX && (frame_full->csub==IAX_TYPE_ACK || frame_full->csub==IAX_TYPE_INVAL)))
					retransmission_new_client(conn->clt, conn, len, (void*)frame_full, now);
			}
		
			if (close_if_received_inval(conn, frame_full, CLOSED_BY_SERVER_INVAL))
				return;

			if(frame_full->type==AST_FRAME_IAX)
				process_packet_closing(conn, frame_full, CLOSED_BY_CLIENT); // check if it's closing
		} else {
			LOG_CONN(NOTICE, "server full frame %i, %i (%s), but call's already closed",
				conn, frame_full->type, frame_full->csub,
				FRAME_NAME(frame_full));
		}
	}
}

static void process_packet_from_server_mini(send_command_t send_command, server *srv, struct iax_packet_mini *frame_mini, size_t len, struct sockaddr_in *addr) {
	connection *conn;

	// check if packet is big enough to analyze - ignore if not
	// mini packet has to have payload
	if(len<=sizeof(struct iax_packet_mini))
		return;

	conn = connection_get_by_srv(srv, 0, ntohs(frame_mini->scallno)&0x7fff);
	if(conn==NULL) {
		LOG(WARNING, "unknown server mini frame from %s", inet_ntoa(addr->sin_addr));

		return; // unknown call - ignore
	} else {
		LOG_CONN(VERBOSE_DEBUG, "server mini frame", conn);
	}

	(void) time(&conn->last_packet);

	if(conn->is_closing==0) {
		ADDRESS_MINI_TO_CLIENT(frame_mini, conn);
		assert(conn->clt!=NULL);
		if(send_command(master_sock, (void*)frame_mini, len, 0, (struct sockaddr*)&conn->clt->addr, sizeof (struct sockaddr_in))<0) {
			err(1, "sending mini frame to client");
		}
	} else {
		// connection's closing and it's minipacket - just drop
	}
}

void process_packet_from_server(send_command_t send_command, server *srv, char *buf, size_t len, struct sockaddr_in *addr, struct timeval *now) {
	struct iax_packet *frame_full = (struct iax_packet *) buf;
	(void) time(&srv->last_activity); // update server response

	if(iax_is_fullframe(frame_full)) { // full frame
		process_packet_from_server_full(send_command, srv, frame_full, len, addr, now);
	} else { // mini frame
		process_packet_from_server_mini(send_command, srv, (struct iax_packet_mini *) buf, len, addr);
	}
}

// fill connection details with some info
static void fill_connection_info_from_ie(connection *conn, struct iax_packet *frame_full, size_t len) {
	ie_position pos;
	
	pos = find_ie(frame_full, len, IE_USERNAME);
	conn->user = strndup((char*)pos.start, pos.length);
	conn->user_len = pos.length;

	pos = find_ie(frame_full, len, IE_CALLING_NUMBER);
	conn->from = strndup((char*)pos.start, pos.length);
	conn->from_len = pos.length;

	pos = find_ie(frame_full, len, IE_CALLED_NUMBER);
	conn->to = strndup((char*)pos.start, pos.length);
	conn->to_len = pos.length;
}

// process the packet incoming from the client side (full frame case)
static void process_packet_from_client_full(send_command_t send_command, client *clt, struct iax_packet *frame_full, size_t len, struct sockaddr_in *addr, struct timeval *now) {
	connection *conn;
	char create;

	// check if packet is big enough to analyze - ignore if not
	if(len<sizeof(struct iax_packet)) {
		LOG(WARNING, "bad packet from %s (too short for IAX) dropped", clt->ip_port);
		return;
	}

	if(frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_ACK) {
		int res = ack_wait_lookup(clt, iax_get_scallno(frame_full), iax_get_dcallno(frame_full), ntohl(frame_full->ts));
		if(res!=0) {
			LOG(NOTICE, "got ACK from %s to our PONG", clt->ip_port);

			// ACKing balancer's PONG - just ignore it and free callno
			client_return_callno(clt, iax_get_dcallno(frame_full));

			client_ref_down_destroy(clt);
			return;
		}
	}

	if(frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_POKE) {
		process_packet_poke(send_command, clt, frame_full, len, addr);
		return;
	}

	// find corresponding connection
	if(frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_NEW)
		create=1;
	else
		create=0;

	conn = connection_get_by_clt(clt, iax_get_scallno(frame_full), iax_get_dcallno(frame_full), create);
	if(conn==NULL) {
		if(create==0) {
			LOG(WARNING, "Got packet type %i, %i (%s) from %s with no connection, scall=%i, dcall=%i, iseq=%i, oseq=%i - dropped",
				frame_full->type, frame_full->csub,
				FRAME_NAME(frame_full),
				clt->ip_port,
				iax_get_scallno(frame_full), iax_get_dcallno(frame_full),
				frame_full->iseqno, frame_full->oseqno);
			// unknown connection && !NEW
			// return INVAL
			packet_create_inval(frame_full);
			if(send_command(master_sock, (void*)frame_full, sizeof (struct iax_packet), 0, (struct sockaddr*) addr, sizeof (struct sockaddr_in))<0) {
				err(1, "sending inval frame back to client");
			}
			// if there are no other connections - drop packet
			client_safe_destroy(clt);
			return;
		} else {
			LOG(ERROR, "Connection from %s creation failed!", clt->ip_port);
			return;
		}
	}

	// check / update the retransmissions
	retransmission_clear(RETR_DEST_CLIENT, frame_full->iseqno - 1, &conn->retransmissions_list);

	if(frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_NEW)
		fill_connection_info_from_ie(conn, frame_full, len);

	(void) time(&conn->last_packet);

	if(conn->is_closing==0 || (frame_full->type==AST_FRAME_IAX && frame_full->csub==IAX_TYPE_ACK)) {
		LOG_CONN(VERBOSE_DEBUG, "client full frame %i, %i (%s) ",
			conn, frame_full->type, frame_full->csub,
			FRAME_NAME(frame_full));

		ADDRESS_FULL_TO_SERVER(frame_full, conn);

		if(send_command(master_sock, (void*)frame_full, len, 0, (struct sockaddr*)&conn->srv->addr, sizeof (struct sockaddr_in))<0) {
			err(1, "sending full frame to server");
		} else {
			if(conn->is_closing==0 && !(frame_full->type==AST_FRAME_IAX && (frame_full->csub==IAX_TYPE_ACK || frame_full->csub==IAX_TYPE_INVAL)))
				retransmission_new_server(conn->srv, conn, len, (void*)frame_full, now);
		}
	} else {
		LOG_CONN(NOTICE, "client full frame %i, %i (%s), but call's already closed",
			conn, frame_full->type, frame_full->csub,
			FRAME_NAME(frame_full));
	}
	
	if (close_if_received_inval(conn, frame_full, CLOSED_BY_CLIENT_INVAL))
		return;

	if(frame_full->type==AST_FRAME_IAX)
		process_packet_closing(conn, frame_full, CLOSED_BY_SERVER); // check if it's closing
}

// process the packet incoming from the client side (mini frame case)
static void process_packet_from_client_mini(send_command_t send_command, client *clt, struct iax_packet_mini *frame_mini, size_t len) {
	connection *conn;

	// check if packet is big enough to analyze - ignore if not
	// mini packet has to have payload
	if(len<=sizeof(struct iax_packet_mini)) {
		client_safe_destroy(clt);
		return;
	}

	conn = connection_get_by_clt(clt, ntohs(frame_mini->scallno)&0x7fff, 0, 0);
	if(conn==NULL) {
		LOG(ERROR, "mini frame from client %s with no connection - dropping", clt->ip_port);
		client_safe_destroy(clt);
		return;
	}

	(void) time(&conn->last_packet);

	if(conn->is_closing==0) {
		ADDRESS_MINI_TO_SERVER(frame_mini, conn);
		if(send_command(master_sock, (void*)frame_mini, len, 0, (struct sockaddr*)&conn->srv->addr, sizeof (struct sockaddr_in))<0) {
			err(1, "sending mini frame to server");
		}
	} else {
		// connection's closing and it's minipacket - just drop
	}
}

// process the packet incoming from the client side
void process_packet_from_client(send_command_t send_command, client *clt, char *buf, size_t len, struct sockaddr_in *addr, struct timeval *now) {
	struct iax_packet *frame_full = (struct iax_packet *) buf;

	if (!clt) {
		LOG(ERROR, "no clt pointer for packet from %s - too many connections?", inet_ntoa(addr->sin_addr));
		packet_create_inval(frame_full);
		if(send_command(master_sock, buf, sizeof (struct iax_packet), 0, (struct sockaddr*) addr, sizeof (struct sockaddr_in))<0) {
			err(1, "sending inval frame to server");
		}
		return;
	}

	if(iax_is_fullframe(frame_full)) {
		process_packet_from_client_full(send_command, clt, frame_full, len, addr, now);
	} else {
		process_packet_from_client_mini(send_command, clt, (struct iax_packet_mini*) buf, len);
	}
}

// analyse the packet
void read_packet(send_command_t send_command, int sock_fd, struct timeval *now) {
	static char buf[65535]; // not reentrant
	ssize_t b, len;
	struct sockaddr_in addr;
	socklen_t addrlen;
	server *srv;

	/* get the packet */
	len = sizeof buf;
	addrlen = sizeof addr;

	b = recvfrom(sock_fd, buf, len, 0, (struct sockaddr *)&addr, &addrlen);
	if (b < 0) {
		warn("recvfrom failed");
		return;
	}

	assert(sizeof addr == addrlen);
	assert(addr.sin_family == AF_INET);

	LOG(VERBOSE_DEBUG, "FD %d is ready, got %ld bytes from %s:%i",
		sock_fd, (long) b, inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

	/* casts to size_t below are safe because we know b is not negative */
	srv = server_get(addr.sin_addr.s_addr);
	if(srv!=NULL) {
		process_packet_from_server(send_command, srv, buf, (size_t)b, &addr, now);
	} else {
		client *clt = client_get(addr.sin_addr, addr.sin_port);
		process_packet_from_client(send_command, clt, buf, (size_t)b, &addr, now);
	}

}
