#include "global.h"
#include "iaxproxy.h"
#include "connections.h"
#include "clients.h"
#include "servers.h"
#include "log.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>

static const char rcsid[] = "iaxproxy-version: " __FILE__ " " VCS_VERSION;

static int daemonize = 1;
static const char *pid_filename = NULL;
static volatile sig_atomic_t stop_counter = 0, stop_now = 0, stop_when_idle = 0;

#ifdef TESTING
#define LOOP_INTERVAL 5
#else
#define LOOP_INTERVAL 60
#endif

static void signal_handler(int sig) {
#define REAL_EXIT_NOW "***\n*** ok, ok, something is wrong, stopping now with abort\n***\n"
#define FLAG_EXIT_NOW "***\n*** got multiple signals, flagging exit WITHOUT waiting for calls to end\n***\n"
#define FLAG_EXIT_WHEN_IDLE "***\n*** got signal, will exit when idle\n***\n"

	fflush(log_file);
	switch(sig) {
		case SIGINT:
		case SIGTERM:
			stop_counter++;

			if (stop_counter >= 6) {
				(void) write(STDOUT_FILENO, REAL_EXIT_NOW, sizeof REAL_EXIT_NOW - 1);
				fflush(log_file);
				fclose(log_file);
				abort();
			}

			if (stop_counter >= 3) {
				(void) write(STDOUT_FILENO, FLAG_EXIT_NOW, sizeof FLAG_EXIT_NOW - 1);
				stop_now++;
			} else {
				(void) write(STDOUT_FILENO, FLAG_EXIT_WHEN_IDLE, sizeof FLAG_EXIT_WHEN_IDLE - 1);
				stop_when_idle++;
			}
			break;

		case SIGUSR1:
			server_list_reload();
			break;

		case SIGHUP:
			open_logs(daemonize);
			break;
	}
}

static void usage(const char *name) {
	printf("Usage:\n");
	printf("%s [-dfhv] [-P port] [-b ip] [-l file] [-p file] [-c file] [-u user/uid] [-s file]\n", name);
	printf("\n");
	printf(" -d       -  one more debug level\n");
	printf(" -f       -  don't fork\n");
	printf(" -h       -  help message\n");
	printf(" -v       -  show version\n");
	printf(" -P port  -  port to bind to (default %d)\n", IAX_PORT);
	printf(" -b ip    -  ip to bind to\n");
	printf(" -l file  -  file for logging\n");
	printf(" -p file  -  pidfile\n");
	printf(" -c file  -  connections log\n");
	printf(" -u user  -  drop root\n");
	printf(" -s file  -  servers list\n");
}

static void spawn_daemon(void) {
	switch(fork()) {
		case 0:	break;
		case -1:
			perror("Can't fork (first)");
			exit(EXIT_FAILURE);
		default:
			exit(EXIT_SUCCESS);
	}

	if(setsid()<0) {
		perror("Can't setsid");
		exit(EXIT_FAILURE);
	}

	switch(fork()) {
		case 0:	break;
		case -1:
			perror("Can't fork (second)");
			exit(EXIT_FAILURE);
		default:
			exit(EXIT_SUCCESS);
	}
}

static void bind_socket(const char *bindaddr, int iax_port) {
	struct sockaddr_in master_addr;

	master_sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (master_sock < 0)
		err(1, "Cannot create listener socket");

	memset(&master_addr, 0, sizeof master_addr);
	master_addr.sin_family = AF_INET;
	master_addr.sin_port = htons(iax_port);
	if(bindaddr==NULL) {
		LOG(INFO, "No IP given - binding to " PROXY_ADDR);
		if(inet_aton(PROXY_ADDR, &master_addr.sin_addr)==0) {
			DIE("Can't set " PROXY_ADDR);
		}
	} else {
		LOG(INFO, "IP given - binding to '%s'", bindaddr);
		if(inet_aton(bindaddr, &master_addr.sin_addr)==0) {
			LOG(ERROR, "Bad IP - binding to 0.0.0.0");
			if(inet_aton(PROXY_ADDR, &master_addr.sin_addr)==0) {
				DIE("Can't set " PROXY_ADDR);
			}
		}
	}
	if (bind(master_sock, (struct sockaddr *)&master_addr, (socklen_t)sizeof master_addr) != 0) {
		int e = errno;
		warn("cannot bind to port");

		if (e == EADDRNOTAVAIL) {
			// - the IP we want isn't mapped, so let's disable this svscan instance
			// we can just run "svc -d" on the current directory because that's where
			// iaxproxy will be running from
			warnx("running svc -d .");
			execlp("svc", "svc", "-d", ".", NULL);
		}

		exit(1);
	}
}

int main(int argc, char **argv) {
	int loop = 0, opt;
	const char *bindaddr = NULL;
	struct timeval tv;
	struct sigaction sa;
	struct timeval now, prev_time;
	int iax_port = IAX_PORT;

	gettimeofday(&now, NULL);

	log_file = stdout;
	while((opt = getopt(argc, argv, "dfhvP:b:l:p:c:u:s:"))!=-1) {
		switch(opt) {
			case 'd':
				debug_level++;
				break;
			case 'h':
			case '?':
				usage(argv[0]);
				exit(EXIT_SUCCESS);
			case 'v':
				printf("%s\n", VERSION);
				exit(EXIT_SUCCESS);
			case 'b':
				bindaddr = optarg;
				break;
			case 'P':
				iax_port = atoi(optarg);
				break;
			case 'l':
				log_filename = optarg;
				break;
			case 'f':
				daemonize = 0;
				break;
			case 'p':
				pid_filename = optarg;
				break;
			case 'c':
				connlog_filename = optarg;
				break;
			case 'u':
				if(atoi(optarg)!=0) {
					if(setuid(atoi(optarg))==-1) {
						err(1, "setuid:");
					}
				} else {
					struct passwd *pw = getpwnam(optarg);
					if(pw==NULL) {
						err(1, "getpwnam - can't find user");
					}
					if(setuid(pw->pw_uid)==-1) {
						err(1, "setuid");
					}
				}
				break;
			case 's':
				servers_list_filename = optarg;
				break;
		}
	}

	if(servers_list_filename == NULL)
		servers_list_filename = "/etc/iaxproxy-servers";

	// catch SIGINT/SIGTERM
	// SIGUSR1 for servers forced reload
	// SIGHUP for cycling log files
	sa.sa_flags = 0;
	sa.sa_handler = signal_handler;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGUSR1, &sa, NULL);
	sigaction(SIGHUP, &sa, NULL);

	if(!log_filename && daemonize) {
		printf("Specify log file if running in daemon mode.\n");
		exit(EXIT_FAILURE);
	}

	open_logs(daemonize);

	if(daemonize) {
		LOG(NOTICE, "Spawning daemon");
		spawn_daemon();
	}
	
	if(pid_filename!=NULL) {
		FILE *pid_file;
		pid_file = fopen(pid_filename, "w");
		if(pid_file == NULL) {
			DIE("Can't write to pid file - exiting.\n");
		}
		fprintf(pid_file, "%i", getpid());
		fclose(pid_file);
	}

	LOG(VERBOSE_DEBUG, "initializing servers list");
	server_list_reload();
	LOG(VERBOSE_DEBUG, "initializing free callno-s");
	free_callno_init();

	// bind to iax_port on chosen / global ip
	bind_socket(bindaddr, iax_port);

	while (stop_now==0) {
		int max = -1, ready;
		fd_set rset;

		FD_ZERO(&rset);

		FD_SET(master_sock, &rset);
		max = master_sock;

		// TODO: change select to simpler wait
		tv.tv_sec = SELECT_TIMEOUT_SEC;
		tv.tv_usec = SELECT_TIMEOUT_USEC;
		ready = select(max + 1, &rset, NULL, NULL, &tv);
//		LOG(VERBOSE_DEBUG, "------------- %d sockets are ready.", ready);

		/* if idle stop is requested, and no active calls, then set the stop now flag */
		if (stop_when_idle!=0) {
			if(connection_check_if_exist() == 0) {
				stop_now = 1;
			}
		}

		if (ready > 0) {
			if (FD_ISSET(master_sock, &rset)) {
				read_packet(&sendto, master_sock, &now);
			} else {
				LOG(ERROR, "select, but not from master socket - bug");
			}
		} else if(ready < 0) {
			if(errno != EINTR)
				warn("select failed");
		}

		prev_time = now;
		(void) gettimeofday(&now, NULL);

		handle_retransmissions(&sendto, master_sock, &now);

		if (now.tv_sec > prev_time.tv_sec) {
			fflush(log_file);
			loop++;
			connection_cleanup_lists(&now);

			check_servers_pokes(&sendto, &now);

			if(loop % LOOP_INTERVAL == 0) { // print server loads once a minute
				LOG_DO(INFO) {
					connection_list_print();
					client_list_print();
				}
				server_list_reload();
				free_callno_log();
			}
		}
	}

	warnx("exiting on signal");

	fflush(log_file);
	{
		// cleanup - optional - just for valgrind test
		connection_destroy_all();
		server_destroy_all();
		client_destroy_all();
	}
	if(connlog_file) {
		fflush(connlog_file);
		fclose(connlog_file);
	}
	fflush(log_file);
	fclose(log_file);

	if(pid_filename!=NULL)
		unlink(pid_filename);

	return (0);
}

