/*
IAX balancing proxy
Authors: Ben Smithrust, Stanislaw Pitucha

Copyright (c) 2007 Gradwell dot com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <sys/stat.h>
#include <sys/types.h>

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "global.h"
#include "servers.h"
#include "log.h"

static const char rcsid[] = "iaxproxy-version: " __FILE__ " " VCS_VERSION;

const char *servers_list_filename=NULL;

struct list_ptrs servers;

/*@null@*/ server* server_get(in_addr_t addr) {
	struct list_ptrs *iter;
	server *srv;

	list_foreach(&servers, iter) {
		srv = list_entry(iter, server, list);
		if(srv->addr.sin_addr.s_addr == addr) {
			return srv;
		}
	}
	return NULL;
}

int server_enable_ip(in_addr_t addr) {
	struct list_ptrs *iter;
	server *srv;

	list_foreach(&servers, iter) {
		srv = list_entry(iter, server, list);
		if(srv->addr.sin_addr.s_addr == addr) {
			srv->enabled = 1;
			return 0;
		}
	}

	srv = (server*)malloc(sizeof(server));
	assert(srv!=NULL);

	srv->enabled=1;
	srv->not_responsive=0;
	srv->load=0;
	srv->last_activity=0;
	srv->last_poke=0;
	srv->last_pong=0;

	memset(&srv->addr, 0, sizeof srv->addr);
	srv->addr.sin_family = AF_INET;
	srv->addr.sin_port = htons(IAX_PORT);
	srv->addr.sin_addr.s_addr = addr;
	srv->ip_string = strdup(inet_ntoa(srv->addr.sin_addr));

	memset(&srv->connections, 0, sizeof (srv->connections));

	list_init(&srv->opening_conns_head);
	list_add_after(&servers, &srv->list);
	return 0;
}

int server_enable(const char* ip) {
	in_addr_t addr = inet_addr(ip);
	if(((int32_t)addr)!=((int32_t)-1))
		return server_enable_ip(addr);
	else
		return -1;
}

int server_disable_ip(in_addr_t addr) {
	struct list_ptrs *iter;
	server *srv;

	list_foreach(&servers, iter) {
		srv = list_entry(iter, server, list);
		if(srv->addr.sin_addr.s_addr == addr) {
			srv->enabled = 0;
			return 0;
		}
	}
	return 1;
}

int server_disable(const char* ip) {
	in_addr_t addr = inet_addr(ip);
	if((int32_t)addr!=-1)
		return server_disable_ip(addr);
	else
		return -1;
}

/*@null@*/ server *server_get_best(void) {
	struct list_ptrs *iter;
	server *srv=NULL;
	server *best_srv=NULL;

	list_foreach(&servers, iter) {
		srv = list_entry(iter, server, list);
		if((srv->enabled!=0) && (srv->not_responsive==0) && (best_srv==NULL || best_srv->load > srv->load)) {
			best_srv = srv;
		}
	}

	return best_srv;
}

void server_list_reload(void) {
	static time_t last_change = 0;
	char buf[4096], *p, *end, *nl;
	int fd, count;
	ssize_t len;
	list_ptrs *iter;
	struct stat servers_list_stat;

	if (stat(servers_list_filename, &servers_list_stat) < 0) {
		warn("cannot run stat on the server list file");
		return;
	}

	if (servers_list_stat.st_mtime <= last_change) {
		return;
	}

	LOG(NOTICE, "server file changed - reloading");
	last_change = servers_list_stat.st_mtime;

	if ((fd = open(servers_list_filename, O_RDONLY)) < 0) {
		warn("cannot open server list %s", SERVER_LIST);
		return;
	}

	len = read(fd, buf, sizeof buf - 1);
	if (len <= 0) {
		warnx("server list %s is empty or unreadable", SERVER_LIST);
		if(close(fd)==-1)
			LOG(ERROR, "close() returned: %s", strerror(errno));
		return;
	}
	if(close(fd)==-1)
		LOG(ERROR, "close() returned: %s", strerror(errno));

	assert(len < (ssize_t)sizeof buf);

	list_foreach(&servers, iter) {
		server *srv = list_entry(iter, server, list);
		srv->enabled=0;
	}

	/* simple line counter */
	end = buf + len;
	*end = '\0';
	p = buf;
	count = 0;
	while (p < end) {
		nl = strchr(p, '\n');
		if (nl) {
			in_addr_t addr;

			*nl = '\0';
			count++;
			addr = inet_addr(p);
			if((int32_t)addr!=-1) {
				(void) server_enable_ip(addr);
			}

			p = nl + 1;

		}
	}

	if (count == 0) {
		warnx("server list %s is corrupt", SERVER_LIST);
		return;
	}

	LOG_DO(INFO) {
		fprintf(log_file, "count = %d\n", count);
		list_foreach(&servers, iter) {
			server *srv = list_entry(iter, server, list);
			if(srv->enabled!=0) {
				if(srv->not_responsive!=0)
					fprintf(log_file, "ip = %s (not responding)\n", srv->ip_string);
				else
					fprintf(log_file, "ip = %s\n", srv->ip_string);
			} else
				fprintf(log_file, "ip = %s (disabled)\n", srv->ip_string);
		}
	}
}

void server_destroy_all(void) {
	list_ptrs *iter;
	int destroyed = 0;

	while((iter = servers.next) != NULL) {
		server *srv = list_entry(iter, server, list);
		list_remove(iter);
		free(srv->ip_string);
		free(srv);
		destroyed++;
	}
	LOG(NOTICE, "destroyed %i servers", destroyed);
}
