#include <stdlib.h>
#include <check.h>
#include "../src/connections.h"
#include "../src/clients.h"
#include "../src/servers.h"
#include "../src/iaxproxy.h"
#include "../src/iax.h"
#include "../src/log.h"

int sent_packets = 0; // counter for the packets sent out

ssize_t verify_send_any(int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen) {
	sent_packets++;
}

START_TEST(test_retrans_create)
{
	struct in_addr address;
	struct timeval now;
	retransmission *ret;

	free_callno_init();
	server_enable("5.6.7.8");

	inet_aton("1.2.3.4", &address);
	client *clt = client_get(address, 1234);
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);

	fail_unless(retransmission_count() == 0, "retransmission list not clean");

	ret = retransmission_new_client(clt, conn, 3, "abc", &now);
	fail_if(ret == NULL, "new retransmission is NULL");
	fail_unless(retransmission_count() == 1, "retransmission not registered");
	
	retransmission_free(ret);
	fail_unless(retransmission_count() == 0, "retransmission not removed");

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

// used by test_retrans_send
ssize_t verify_send_abc_to_1234(int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen) {
	struct in_addr address;
	const char *addr = inet_ntoa(((const struct sockaddr_in*)dest_addr)->sin_addr);

	fail_unless(iax_is_retransmission((struct iax_packet*) buf));
	fail_unless(strcmp("1.2.3.4", addr) == 0, "address not matching");
	fail_unless(len == sizeof(struct iax_packet)+3, "len not matching");
	fail_unless(memcmp(&((struct iax_packet*)buf)->iedata, "abc", 3) == 0, "data not matching");

	sent_packets++;
}

START_TEST(test_retrans_send)
{
	struct in_addr address;
	struct timeval now;
	retransmission *ret;
	struct {struct iax_packet pkt; unsigned char data[3];} pkt = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000), .ts = 0,
			.oseqno = 0, .iseqno = 0, .type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			'a', 'b', 'c'
		}
	};

	free_callno_init();
	server_enable("5.6.7.8");

	inet_aton("1.2.3.4", &address);
	client *clt = client_get(address, 1234);
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);
	
	ret = retransmission_new_client(clt, conn, sizeof(pkt), (const char*)&pkt, &now);

	fail_unless(ret->no_tries == 0, "number of tries not zeroed");

	handle_retransmissions(NULL, 0, &now);
	fail_unless(ret->no_tries == 0, "sending should not be tried yet");

	now.tv_sec += 1;
	handle_retransmissions(&verify_send_abc_to_1234, 0, &now);
	fail_unless(ret->no_tries == 1, "number of tries not updated");
	
	retransmission_free(ret);
	connection_destroy_all();
	server_destroy_all();
}
END_TEST

START_TEST(test_retrans_send_repeat_quickly)
{
	struct in_addr address;
	struct timeval now;
	retransmission *ret;
	struct {struct iax_packet pkt; unsigned char data[3];} pkt = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000), .ts = 0,
			.oseqno = 0, .iseqno = 0, .type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			'a', 'b', 'c'
		}
	};

	free_callno_init();
	server_enable("5.6.7.8");

	inet_aton("1.2.3.4", &address);
	client *clt = client_get(address, 1234);
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);
	
	ret = retransmission_new_client(clt, conn, sizeof(pkt), (const char*)&pkt, &now);

	fail_unless(ret->no_tries == 0, "number of tries not zeroed");

	now.tv_sec += 1;
	handle_retransmissions(&verify_send_abc_to_1234, 0, &now);
	handle_retransmissions(&verify_send_any, 0, &now);

	fail_unless(ret->no_tries == 1, "packet should be retransmitted once only");
	
	retransmission_free(ret);
	connection_destroy_all();
	server_destroy_all();
}
END_TEST

START_TEST(test_retrans_send_max)
{
	struct in_addr address;
	struct timeval now;
	retransmission *ret;
	struct {struct iax_packet pkt; unsigned char data[3];} pkt = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000), .ts = 0,
			.oseqno = 0, .iseqno = 0, .type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			'a', 'b', 'c'
		}
	};
	
	free_callno_init();
	server_enable("5.6.7.8");

	inet_aton("1.2.3.4", &address);
	client *clt = client_get(address, 1234);
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);
	
	ret = retransmission_new_client(clt, conn, sizeof(pkt), (const char*)&pkt, &now);

	fail_unless(ret->no_tries == 0, "number of tries not zeroed");

	now.tv_sec += 1;
	handle_retransmissions(&verify_send_abc_to_1234, 0, &now);
	
	now.tv_sec += 1;
	handle_retransmissions(&verify_send_abc_to_1234, 0, &now);
	
	now.tv_sec += 1;
	handle_retransmissions(&verify_send_abc_to_1234, 0, &now);

	fail_unless(ret->no_tries == 3, "number of tries not updated");
	
	now.tv_sec += 1;
	handle_retransmissions(&verify_send_abc_to_1234, 0, &now);
	
	fail_unless(retransmission_count() == 0, "retransmission not removed");
	
	connection_destroy_all();
	server_destroy_all();
}
END_TEST

START_TEST(test_retrans_clt_ref)
{
	struct in_addr address;
	struct timeval now;
	retransmission *ret;
	
	free_callno_init();
	server_enable("5.6.7.8");

	inet_aton("1.2.3.4", &address);
	client *clt = client_get(address, 1234);
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);
	
	fail_unless(clt->ref_count == 1, "client should be referenced by connection only");

	ret = retransmission_new_client(clt, conn, 3, "abc", &now);
	fail_unless(clt->ref_count == 2, "client should be referenced by connection and retransmission");
	
	retransmission_free(ret);
	fail_unless(clt->ref_count == 1, "client should be referenced by connection only, again");
	
	connection_destroy_all();
	server_destroy_all();
}
END_TEST

ssize_t verify_send_new_to_5678(int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen) {
	struct in_addr address;
	const char *addr = inet_ntoa(((const struct sockaddr_in*)dest_addr)->sin_addr);
	const struct iax_packet *pkt = buf;

	fail_unless(strcmp("5.6.7.8", addr) == 0, "address not matching");
	fail_unless(len == sizeof(struct iax_packet)+2+6+2+6+2+6, "len not matching (3 IE x 6 chars)");
	fail_unless(pkt->type == AST_FRAME_IAX, "frame type changed");
	fail_unless(pkt->csub == IAX_TYPE_NEW, "iax type changed");
	
	sent_packets++;
}

START_TEST(test_packet_new)
{
	struct sockaddr_in address;
	struct timeval now;
	retransmission *ret;
	struct {struct iax_packet pkt; unsigned char data[24];} pkt = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000),
			.ts = 0,
			.oseqno = 0, .iseqno = 0,
			.type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			IE_CALLED_NUMBER, 6,
			'1', '2', '3', '4', '5', '6',
			IE_CALLING_NUMBER, 6,
			'2', '3', '4', '5', '6', '7',
			IE_USERNAME, 6,
			'a', 'b', 'c', 'd', 'e', 'f',
		}
	};
	
	free_callno_init();
	server_enable("5.6.7.8");

	inet_aton("1.2.3.4", &address.sin_addr);
	client *clt = client_get(address.sin_addr, 1234);
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);
	
	fail_unless(sent_packets == 0, "packets sent before test?");
	fail_unless(retransmission_count() == 0, "retransmission list not clean");
	process_packet_from_client(&verify_send_new_to_5678, clt, (void*) &pkt, sizeof(struct iax_packet)+((2+6)*3), &address, &now);
	fail_unless(sent_packets == 1, "packet not sent");
	fail_unless(retransmission_count() == 1, "retransmission not set up (currently: %i)", retransmission_count());
	fail_unless(retransmission_count_connection(conn) == 1, "retransmission not set up in the connection (currently: %i)", retransmission_count_connection(conn));

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

ssize_t verify_send_accept_to_1234(int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen) {
	struct in_addr address;
	const char *addr = inet_ntoa(((const struct sockaddr_in*)dest_addr)->sin_addr);
	const struct iax_packet *pkt = buf;

	fail_unless(strcmp("1.2.3.4", addr) == 0, "address not matching");
	fail_unless(len == sizeof(struct iax_packet)+6, "len not matching (1 IE_FORMAT)");
	fail_unless(pkt->type == AST_FRAME_IAX, "frame type changed");
	fail_unless(pkt->csub == IAX_TYPE_ACCEPT, "iax type changed");
	
	sent_packets++;
}

START_TEST(test_packet_retransmission_clearing)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[24];} pkt_new= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000),
			.ts = 0,
			.oseqno = 0, .iseqno = 0,
			.type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			IE_CALLED_NUMBER, 6,
			'1', '2', '3', '4', '5', '6',
			IE_CALLING_NUMBER, 6,
			'2', '3', '4', '5', '6', '7',
			IE_USERNAME, 6,
			'a', 'b', 'c', 'd', 'e', 'f',
		}
	};
	struct {struct iax_packet pkt; unsigned char data[24];} pkt_accept = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001),
			.ts = 100,
			.oseqno = 0, .iseqno = 1,
			.type = AST_FRAME_IAX, .csub = IAX_TYPE_ACCEPT
		},
		.data = {
			IE_FORMAT, 4,
			'0', '0', '0', '0',
		}
	};
	
	free_callno_init();

	inet_aton("5.6.7.8", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("1.2.3.4", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);

	fail_unless(clt->ref_count == 1, "only the connection should reference the client");
	fail_unless(sent_packets == 0, "packets sent before test?");
	fail_unless(retransmission_count() == 0, "retransmission list not clean");
	process_packet_from_client(&verify_send_new_to_5678, clt, (void*) &pkt_new, sizeof(struct iax_packet)+((2+6)*3), &address_client, &now);
	fail_unless(clt->ref_count == 1, "connection and retransmission should reference the client");
	fail_unless(sent_packets == 1, "packet NEW not sent");
	fail_unless(retransmission_count() == 1, "retransmission not set up (currently: %i)", retransmission_count());
	fail_unless(retransmission_count_connection(conn) == 1, "retransmission not set up in the connection (currently: %i)", retransmission_count_connection(conn));
	
	pkt_accept.pkt.dcallno = htons(conn->balancer_cltno);

	process_packet_from_server(&verify_send_accept_to_1234, srv, (void*) &pkt_accept, sizeof(struct iax_packet)+6, &address_server, &now);
	fail_unless(sent_packets == 2, "packet ACCEPT not sent");
	fail_unless(retransmission_count() == 1, "retransmission not removed (currently: %i)", retransmission_count());
	fail_unless(retransmission_count_connection(conn) == 1, "retransmission not removed from the connection (currently: %i)", retransmission_count_connection(conn));
	fail_unless(clt->ref_count == 2, "connection and retransmission should reference the client");

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

START_TEST(test_packet_retransmission_to_server)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[24];} pkt_new= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000),
			.ts = 0,
			.oseqno = 0, .iseqno = 0,
			.type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			IE_CALLED_NUMBER, 6,
			'1', '2', '3', '4', '5', '6',
			IE_CALLING_NUMBER, 6,
			'2', '3', '4', '5', '6', '7',
			IE_USERNAME, 6,
			'a', 'b', 'c', 'd', 'e', 'f',
		}
	};
	
	free_callno_init();

	inet_aton("5.6.7.8", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("1.2.3.4", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);

	fail_unless(sent_packets == 0, "packets sent before test?");
	fail_unless(retransmission_count() == 0, "retransmission list not clean");
	process_packet_from_client(&verify_send_new_to_5678, clt, (void*) &pkt_new, sizeof(struct iax_packet)+((2+6)*3), &address_client, &now);
	fail_unless(sent_packets == 1, "packet NEW not sent");
	fail_unless(retransmission_count_connection(conn) == 1, "retransmission not set up in the connection (currently: %i)", retransmission_count_connection(conn));

	now.tv_sec += 1;
	handle_retransmissions(&verify_send_new_to_5678, 0, &now);
	fail_unless(sent_packets == 2, "packet NEW not resent");

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

START_TEST(test_packet_no_retransmission_duplicate)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[24];} pkt_new= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000),
			.ts = 0,
			.oseqno = 0, .iseqno = 0,
			.type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			IE_CALLED_NUMBER, 6,
			'1', '2', '3', '4', '5', '6',
			IE_CALLING_NUMBER, 6,
			'2', '3', '4', '5', '6', '7',
			IE_USERNAME, 6,
			'a', 'b', 'c', 'd', 'e', 'f',
		}
	};
	
	free_callno_init();

	inet_aton("5.6.7.8", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("1.2.3.4", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));
	connection *conn = connection_get_by_clt(clt, 1, 1, 1);

	gettimeofday(&now, NULL);

	process_packet_from_client(&verify_send_new_to_5678, clt, (void*) &pkt_new, sizeof(struct iax_packet)+((2+6)*3), &address_client, &now);
	iax_set_retransmission((struct iax_packet*) &pkt_new, 1);

	process_packet_from_client(&verify_send_new_to_5678, clt, (void*) &pkt_new, sizeof(struct iax_packet)+((2+6)*3), &address_client, &now);
	fail_unless(sent_packets == 2, "both NEW packets should be sent");
	fail_unless(retransmission_count_connection(conn) == 1, "retransmission should be set up once only (currently: %i)", retransmission_count_connection(conn));

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

START_TEST(test_connection_inval)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[24];} pkt_new= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000), .ts = 0,
			.oseqno = 0, .iseqno = 0, .type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			IE_CALLED_NUMBER, 6, '1', '2', '3', '4', '5', '6',
			IE_CALLING_NUMBER, 6, '2', '3', '4', '5', '6', '7',
			IE_USERNAME, 6, 'a', 'b', 'c', 'd', 'e', 'f',
		}
	};
	
	struct {struct iax_packet pkt; unsigned char data[6];} pkt_accept= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(10),
			.oseqno = 0, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_ACCEPT
		},
		.data = {
			IE_FORMAT, 4, '0', '0', '0', '0',
		}
	};
	
	struct {struct iax_packet pkt; unsigned char data[0];} pkt_inval= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(10),
			.oseqno = 1, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_INVAL
		},
		.data = {
		}
	};
	
	free_callno_init();

	inet_aton("5.6.7.8", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("1.2.3.4", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));

	gettimeofday(&now, NULL);

	process_packet_from_client(&verify_send_any, clt, (void*) &pkt_new, sizeof(pkt_new), &address_client, &now);
	fail_unless(sent_packets == 1, "packet NEW not sent");
	fail_unless(connection_count() == 1, "one connection expected after NEW");
	
	connection *conn = connection_get_by_clt(clt, 1, 1, 0);
	fail_unless(conn != NULL, "connection not found");
	pkt_accept.pkt.dcallno = pkt_inval.pkt.dcallno = htons(0x8000 | (0x7fff & conn->balancer_cltno));

	process_packet_from_server(&verify_send_any, srv, (void*) &pkt_accept, sizeof(pkt_accept), &address_server, &now);
	fail_unless(sent_packets == 2, "packet ACCEPT not sent");
	fail_unless(connection_count() == 1, "one connection expected after ACCEPT");
	
	process_packet_from_client(&verify_send_any, clt, (void*) &pkt_inval, sizeof(pkt_inval), &address_client, &now);
	fail_unless(connection_count() == 0, "expected the connections to finish");
	fail_unless(retransmission_count() == 0, "expected no retransmissions after the call finished (got %i)", retransmission_count());
	fail_unless(sent_packets == 3, "packet INVALID not sent");

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

ssize_t verify_send_1234_inval(int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen) {
	struct in_addr address;
	const char *addr = inet_ntoa(((const struct sockaddr_in*)dest_addr)->sin_addr);
	const struct iax_packet *pkt = buf;

	fail_unless(strcmp("1.2.3.4", addr) == 0, "address not matching");
	fail_unless(len == sizeof(struct iax_packet), "len not matching (should be no data)");
	fail_unless(pkt->type == AST_FRAME_IAX, "frame type not matching");
	fail_unless(pkt->csub == IAX_TYPE_INVAL, "iax type not matching");
	
	sent_packets++;
}

START_TEST(test_connection_unknown_client)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[6];} pkt_accept= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(10),
			.oseqno = 0, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_ACCEPT
		},
		.data = {
			IE_FORMAT, 4, '0', '0', '0', '0',
		}
	};
	
	free_callno_init();

	inet_aton("5.6.7.8", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("1.2.3.4", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));

	gettimeofday(&now, NULL);

	process_packet_from_client(&verify_send_1234_inval, clt, (void*) &pkt_accept, sizeof(pkt_accept), &address_client, &now);

	fail_unless(sent_packets == 1, "response not sent");
	fail_unless(connection_count() == 0, "no connections expected");
	fail_unless(retransmission_count() == 0, "expected no retransmissions (got %i)", retransmission_count());

	server_destroy_all();
}
END_TEST

START_TEST(test_connection_unknown_server)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[6];} pkt_accept= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(10),
			.oseqno = 0, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_ACCEPT
		},
		.data = {
			IE_FORMAT, 4, '0', '0', '0', '0',
		}
	};
	
	free_callno_init();

	inet_aton("1.2.3.4", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("5.6.7.8", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));

	gettimeofday(&now, NULL);

	process_packet_from_server(&verify_send_1234_inval, srv, (void*) &pkt_accept, sizeof(pkt_accept), &address_server, &now);

	fail_unless(sent_packets == 1, "response not sent");
	fail_unless(connection_count() == 0, "no connections expected");
	fail_unless(retransmission_count() == 0, "expected no retransmissions (got %i)", retransmission_count());

	server_destroy_all();
}
END_TEST

START_TEST(test_connection_closing)
{
	struct sockaddr_in address_server, address_client;
	struct timeval now;
	retransmission *ret;
	int balancer_callno;

	struct {struct iax_packet pkt; unsigned char data[24];} pkt_new= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0000), .ts = 0,
			.oseqno = 0, .iseqno = 0, .type = AST_FRAME_IAX, .csub = IAX_TYPE_NEW
		},
		.data = {
			IE_CALLED_NUMBER, 6, '1', '2', '3', '4', '5', '6',
			IE_CALLING_NUMBER, 6, '2', '3', '4', '5', '6', '7',
			IE_USERNAME, 6, 'a', 'b', 'c', 'd', 'e', 'f',
		}
	};
	
	struct {struct iax_packet pkt; unsigned char data[6];} pkt_accept= {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(10),
			.oseqno = 0, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_ACCEPT
		},
		.data = {
			IE_FORMAT, 4, '0', '0', '0', '0',
		}
	};
	
	struct {struct iax_packet pkt; unsigned char data[0];} pkt_ack = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(10),
			.oseqno = 1, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_ACK
		}, .data = { }
	};
	
	struct {struct iax_packet pkt; unsigned char data[0];} pkt_hangup = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(20),
			.oseqno = 1, .iseqno = 1, .type = AST_FRAME_IAX, .csub = IAX_TYPE_HANGUP
		}, .data = { }
	};
	
	struct {struct iax_packet pkt; unsigned char data[0];} pkt_hangup_ack = {
		.pkt = {
			.scallno = htons(0x8001), .dcallno = htons(0x0001), .ts = htonl(20),
			.oseqno = 1, .iseqno = 2, .type = AST_FRAME_IAX, .csub = IAX_TYPE_ACK
		}, .data = { }
	};

	free_callno_init();

	inet_aton("5.6.7.8", &address_server.sin_addr);
	address_server.sin_port = IAX_PORT;
	inet_aton("1.2.3.4", &address_client.sin_addr);
	address_client.sin_port = IAX_PORT;
	
	server_enable_ip(address_server.sin_addr.s_addr);
	server *srv = server_get(address_server.sin_addr.s_addr);

	client *clt = client_get(address_client.sin_addr, htons(IAX_PORT));

	gettimeofday(&now, NULL);

	fail_unless(sent_packets == 0, "packets sent before test?");
	fail_unless(connection_count() == 0, "no connections expected");
	process_packet_from_client(&verify_send_any, clt, (void*) &pkt_new, sizeof(pkt_new), &address_client, &now);
	fail_unless(sent_packets == 1, "packet NEW not sent");
	fail_unless(connection_count() == 1, "one connection expected after NEW");
	fail_unless(retransmission_count() == 1, "expected 1 retransmission (NEW) (got %i)", retransmission_count());
	
	connection *conn = connection_get_by_clt(clt, 1, 1, 0);
	fail_unless(conn != NULL, "connection not found");
	pkt_accept.pkt.dcallno = pkt_ack.pkt.dcallno = pkt_hangup.pkt.dcallno = pkt_hangup_ack.pkt.dcallno =
		htons(0x8000 | (0x7fff & conn->balancer_cltno));
	
	process_packet_from_server(&verify_send_any, srv, (void*) &pkt_accept, sizeof(pkt_accept), &address_server, &now);
	fail_unless(sent_packets == 2, "packet ACCEPT not sent");
	fail_unless(retransmission_count() == 1, "expected 1 retransmission (ACCEPT) (got %i)", retransmission_count());

	process_packet_from_client(&verify_send_any, clt, (void*) &pkt_ack, sizeof(pkt_ack), &address_client, &now);
	fail_unless(sent_packets == 3, "packet ACK(accept) not sent");
	fail_unless(retransmission_count() == 0, "expected 0 retransmissions after ACK(accept) (got %i)", retransmission_count());
	
	process_packet_from_client(&verify_send_any, clt, (void*) &pkt_hangup, sizeof(pkt_hangup), &address_client, &now);
	fail_unless(sent_packets == 4, "packet HANGUP not sent");
	fail_unless(retransmission_count() == 1, "expected 1 retransmission (HANGUP) (got %i)", retransmission_count());
	
	process_packet_from_server(&verify_send_any, srv, (void*) &pkt_hangup_ack, sizeof(pkt_hangup_ack), &address_server, &now);
	fail_unless(sent_packets == 5, "packet ACK(hangup) not sent");
	
	fail_unless(connection_count() == 0, "expected the connections to finish");
	fail_unless(retransmission_count() == 0, "expected no retransmissions after the call finished (got %i)", retransmission_count());

	connection_destroy_all();
	server_destroy_all();
}
END_TEST

Suite *iaxproxy_suite (void) {
	Suite *s = suite_create ("Connections");

	/* Core test case */
	TCase *tc_retrans = tcase_create ("retransmissions");
	tcase_add_test (tc_retrans, test_retrans_create);
	tcase_add_test (tc_retrans, test_retrans_send);
	tcase_add_test (tc_retrans, test_retrans_send_repeat_quickly);
	tcase_add_test (tc_retrans, test_retrans_send_max);
	tcase_add_test (tc_retrans, test_retrans_clt_ref);
	suite_add_tcase (s, tc_retrans);
	
	TCase *tc_packet = tcase_create ("packet handling");
	tcase_add_test (tc_packet, test_packet_new);
	tcase_add_test (tc_packet, test_packet_retransmission_clearing);
	tcase_add_test (tc_packet, test_packet_retransmission_to_server);
	tcase_add_test (tc_packet, test_packet_no_retransmission_duplicate);
	suite_add_tcase (s, tc_packet);

	TCase *tc_connection = tcase_create ("connections");
	tcase_add_test (tc_connection, test_connection_closing);
	tcase_add_test (tc_connection, test_connection_inval);
	tcase_add_test (tc_connection, test_connection_unknown_client);
	tcase_add_test (tc_connection, test_connection_unknown_server);
	suite_add_tcase (s, tc_connection);

	return s;
}
		
int main (void) {
	int number_failed;
	Suite *s = iaxproxy_suite ();
	SRunner *sr = srunner_create (s);
	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);
	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

